// Copyright (C) 2021-2022 David Sugar <tychosoft@gmail.com>.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package main

import (
	"fmt"
	"os"
	"os/signal"
	"runtime"
	"syscall"

	"github.com/alexflint/go-arg"
	"github.com/gofiber/fiber/v2"

	"storehouse/lib"
)

// Argument parser....
type Args struct {
	Host    string `arg:"env:STOREHOUSE_HOST,--host" help:"server host address" default:""`
	Port    uint16 `arg:"env:STOREHOUSE_PORT,--port" help:"server port" default:"8080"`
	Prefix  string `arg:"--prefix" help:"server prefix path"`
	Verbose int    `arg:"-v" help:"debugging log level"`
}

var (
	// binding configs
	prefixPath = "/var/lib/storehouse"
	logPrefix  = "/var/log"
	version    = "unknown"

	// globals
	args *Args = &Args{Prefix: prefixPath}
)

func (Args) Version() string {
	return "Version: " + version
}

func (Args) Description() string {
	return "storehouse - provides storehouse access daemon"
}

func init() {
	for pos, arg := range os.Args {
		switch arg {
		case "--":
			return
		case "-vv":
			os.Args[pos] = "--verbose=2"
		case "-vvv":
			os.Args[pos] = "--verbose=3"
		case "-vvvv":
			os.Args[pos] = "--verbose=4"
		case "-vvvvv":
			os.Args[pos] = "--verbose=5"
		}
	}
}

func main() {
	logPath := logPrefix + "/storehouse.log"
	arg.MustParse(args)

	// setup services
	lib.Logger(args.Verbose, logPath)
	lib.Debug(3, "working directory ", args.Prefix)
	err := os.Chdir(args.Prefix)
	if err != nil {
		lib.Fail(1, err)
	}

	// config app
	address := fmt.Sprintf("%s:%v", args.Host, args.Port)
	app := fiber.New(fiber.Config{
		ServerHeader:          "Storehouse",
		AppName:               "Storehouse v" + version,
		GETOnly:               true,
		DisableStartupMessage: !lib.DebugFlag || args.Verbose < 1,
		StreamRequestBody:     true,
	})

	// routes, etc...
	app.Static("/", "./tarballs", fiber.Static{Browse: true})
	app.Static("/tarballs", "./tarballs", fiber.Static{Browse: true})
	app.Static("/archive", "./archve")
	app.Static("/alpine", "./alpine")
	app.Static("/data", "./data")
	app.Static("/docs", "./docs", fiber.Static{Browse: true})
	app.Static("/images", "./images")
	app.Static("/screenshots", "./screenshots")

	// signal handler...
	signals := make(chan os.Signal, 1)
	signal.Notify(signals, os.Interrupt, syscall.SIGTERM, syscall.SIGHUP)
	go func() {
		defer app.Shutdown()
		for {
			switch <-signals {
			case os.Interrupt: // sigint/ctrl-c
				fmt.Println()
				return
			case syscall.SIGTERM: // normal exit
				return
			case syscall.SIGHUP: // cleanup
				lib.Info("reload service")
				lib.LoggerRestart()
				runtime.GC()
			}
		}
	}()

	// start service(s)...
	lib.Info("start service on ", address)
	if err := app.Listen(address); err != nil {
		lib.Fail(99, err)
	}
	lib.Info("stop service")
}
