// Copyright (C) 2021-2022 David Sugar <tychosoft@gmail.com>.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package lib

import (
	"encoding/json"
	"errors"
	"time"
)

// lib.Time reports utc and truncates to nearest second
type Time time.Time

// get current time
func Now() Time {
	return Time(time.Now().UTC().Truncate(time.Second))
}

// marshall as duration string (zulu time) nearest second
func (t Time) MarshalJSON() ([]byte, error) {
	return json.Marshal(time.Time(t).UTC().Truncate(time.Second).String())
}

// Time as zulu time string
func (t Time) String() string {
	return time.Time(t).UTC().Truncate(time.Second).String()
}

// Local time format such as for logs
func (t Time) Local() string {
	return time.Time(t).Local().Truncate(time.Second).Format("2001-03-28 15:03:04")
}

// Parse Time
func (t *Time) Parse(text string) (Time, error) {
	tmp, err := time.Parse(time.RFC3339, text)
	return Time(tmp.Truncate(time.Second)), err
}

// unmarshall from float or string form
func (t *Time) UnmarshalJSON(b []byte) error {
	var v interface{}
	if err := json.Unmarshal(b, &v); err != nil {
		return err
	}
	switch value := v.(type) {
	case string:
		tmp, err := time.Parse(time.RFC3339, value)
		if err != nil {
			return err
		}
		*t = Time(tmp.Truncate(time.Second))
		return nil
	default:
		return errors.New("invalid duration")
	}
}
